import re
import functools
import time
import collections
import os
import numpy as np

try:
    import torch
except ImportError:
    pass

from ._model import Chat
from ._local import Local


class LocalMock(Local):
    def __init__(self, byte_patterns=[], echo=True):
        
        super().__init__(
            # our tokens are all bytes and all lowercase letter pairs
            [b"<s>"] + [bytes([i,j]) for i in range(ord('a'), ord('z')) for j in range(ord('a'), ord('z'))] + [bytes([i]) for i in range(256)],
            0,
            0,
            echo=echo
        )

        self._valid_mask = np.zeros(len(self.tokens))
        for i,t in enumerate(self.tokens):
            try:
                t.decode("utf8")
                self._valid_mask[i] = 1.0
            except:
                pass

        # allow a single byte pattern to be passed
        if isinstance(byte_patterns, (bytes, str)):
            byte_patterns = [byte_patterns]

        # allow for strings to be passed
        for i,pattern in enumerate(byte_patterns):
            if isinstance(pattern, str):
                byte_patterns[i] = pattern.encode("utf8")

        self.byte_patterns = byte_patterns
        self._rand_generator = np.random.default_rng(seed=42)

        self._cache_state["past_key_values"] = None
        self._cache_state["logits"] = None
        self.eos_token = b"<s>"

    def _get_logits(self, token_ids, forced_bytes):
        '''Pretends to compute the logits for the given token state.
        '''

        # build the byte strings
        byte_string = b"".join(self.tokens[i] for i in token_ids)

        # we randomly generate valid unicode bytes
        logits = self._rand_generator.standard_normal(len(self.tokens)) * self._valid_mask

        # if we have a pattern that matches then force the next token
        bias = 100.0
        if self.byte_patterns is not None:
            byte_string
            for p in self.byte_patterns:
                if p.startswith(byte_string) and len(p) > len(byte_string):
                    for i in self._get_next_tokens(p[len(byte_string):]):
                        logits[i] += bias
                    bias /= 2 # if we have multiple matches then they apply with decreasing bias
        
        return torch.tensor(logits)
    
    def _get_next_tokens(self, byte_string):
        for i,t in enumerate(self.tokens):
            if byte_string.startswith(t):
                yield i
        

class LocalMockChat(LocalMock, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


from guidance import models

class ContextBlock:
    def __init__(self, opener, closer):
        self.opener = opener
        self.closer = closer

    def __enter__(self):
        models.Model._open_blocks[self] = None
    
    def __exit__(self, exc_type, exc_value, traceback):
        del models.Model._open_blocks[self]

def block(name=None, opener="", closer="", hidden=False):
    assert name is None
    assert hidden is False
    return ContextBlock(opener, closer)


import os
import requests
import inspect
import json
import re
import asyncio
import queue
import ast
import types
import itertools
import textwrap
import sys

class TextRange:
    def __init__(self, start, end, lm):
        self.start = start
        self.end = end
        self.lm = lm

    def __str__(self) -> str:
        return str(self.lm)[self.start:self.end]

# class InPlace():
#     """Creates a scope where the LM is in-place or not."""
#     def __init__(self, lm):
#         self.lm = lm
#         self._prev_inplace = lm._inplace
    
#     def __enter__(self):
#         new_lm = self.lm._clone()
#         new_lm._inplace = True
#         return new_lm

#     def __exit__(self, type, value, traceback):
#         InPlace._rec_set_inplace(self.lm, self._prev_inplace)

#     @staticmethod
#     def _rec_set_inplace(lm, value):
#         lm._inplace = value
#         for child in lm._children:
#             InPlace._rec_set_inplace(child, value)

class Silent():
    """Creates a scope where the LM is silent or not."""
    def __init__(self, lm, silent):
        self.lm = lm
        self._prev_silent = lm._silent
        self.silent = silent
    
    def __enter__(self):
        if self.silent is not None:
            self.lm._silent = self.silent

    def __exit__(self, type, value, traceback):
        if self.silent is not None:
            Silent._rec_set_silent(self.lm, self._prev_silent)

    @staticmethod
    def _rec_set_silent(lm, value):
        lm._silent = value
        for child in lm._children:
            Silent._rec_set_silent(child, value)

class Hidden():
    """Creates a scope where the LM state is optionally hidden from following calls.
    
    Hidden means that the text inside this scope will not be used as context for
    later calls.
    """
    def __init__(self, lm, hidden):
        self.lm = lm
        self.hidden = hidden
    
    def __enter__(self):
        self.offset = len(self.lm)

    def __exit__(self, type, value, traceback):
        if self.hidden:
            Hidden._rec_make_hidden(self.lm, self.offset)

    @staticmethod
    def _rec_make_hidden(lm, offset):
        lm.reset(offset, clear_variables=False)
        for child in lm._children:
            Hidden._rec_make_hidden(child, offset)

class _Rewrite(ast.NodeTransformer):
    def visit_Constant(self, node):
        # print(node)
        if isinstance(node.value, str) and node.lineno < node.end_lineno:
            self.start_counts[node.lineno-1] += 1
            start_line = self.source_lines[node.lineno-1]
            start_string = start_line[node.col_offset:]
            
            # check for literal multiline strings
            if start_string.startswith("f'''") or start_string.startswith("'''") or start_string.startswith('f"""') or start_string.startswith('"""'):
                
                # track our indentation level
                if self.indentation[node.lineno-1] is None:
                    indent = start_line[:len(start_line) - len(start_line.lstrip())]
                    for i in range(node.lineno-1, node.end_lineno):
                        self.indentation[i] = indent
                indent = self.indentation[node.lineno-1]

                # strip indentation when it is consistent
                lines = node.value.split("\n")
                fail = False
                new_lines = []
                for i,line in enumerate(lines):
                    if (i == 0 and (self.start_counts[node.lineno-1] > 1 or not start_line.endswith("\\"))) or line == "":
                        new_lines.append(line)
                    elif line.startswith(indent):
                        new_lines.append(line[len(indent):])
                    # elif (i == 0 and line.endswith("\\")) or line == "":
                    #     new_lines.append(line)
                    else:
                        fail = True
                        break
                if not fail:
                    node.value = "\n".join(new_lines)

        return node
class normalize_notebook_stdout_stderr():
    '''Remaps stdout and stderr back to their normal selves from what ipykernel did to them.
    
    Based on: https://github.com/ipython/ipykernel/issues/795
    '''

    def __enter__(self):
        normal_stdout = sys.__stdout__.fileno()
        self.restore_stdout = None
        if getattr(sys.stdout, "_original_stdstream_copy", normal_stdout) != normal_stdout:
            self.restore_stdout = sys.stdout._original_stdstream_copy
            sys.stdout._original_stdstream_copy = normal_stdout

        normal_stderr = sys.__stderr__.fileno()
        self.restore_stderr = None
        if getattr(sys.stderr, "_original_stdstream_copy", normal_stderr) != normal_stderr:
            self.restore_stderr = sys.stderr._original_stdstream_copy
            sys.stderr._original_stdstream_copy = normal_stderr

    def __exit__(self, exc_type, exc_value, traceback):
        if self.restore_stdout is not None:
            sys.stderr._original_stdstream_copy = self.restore_stdout
        if self.restore_stderr is not None:
            sys.stderr._original_stdstream_copy = self.restore_stderr

def strip_multiline_string_indents(f):

    source = textwrap.dedent(inspect.getsource(f))
    blanks = '\n' * f.__code__.co_firstlineno # padd the source so the lines in the file line up for the debugger
    source = blanks + '\n'.join(source.splitlines()[1:]) # remove the decorator first line.
    # print(source)

    old_code_obj = f.__code__
    old_ast = ast.parse(source)
    r = _Rewrite()
    r.source_lines = source.split("\n")
    r.indentation = [None for l in r.source_lines]
    r.start_counts = [0 for l in r.source_lines]
    # r._avoid_backslashes = True
    new_ast = r.visit(old_ast)
    new_code_obj = compile(new_ast, old_code_obj.co_filename, 'exec')

    # find the code block
    for i in range(len(new_code_obj.co_consts)):
        if str(type(new_code_obj.co_consts[i])) == "<class 'code'>":
            break

    # create a new function based on the modified code
    new_f = types.FunctionType(
        new_code_obj.co_consts[i],
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__
    )
    new_f.__kwdefaults__ = f.__kwdefaults__
    return new_f

class CaptureEvents():
    """Creates a scope where all the events are captured in a queue.
    
    Note that this does not stop the events from being captured by higher level scopes.
    """
    def __init__(self, lm):
        self.lm = lm
    
    def __enter__(self):
        self.lm._event_queue = queue.Queue()
        return self.lm._event_queue

    def __exit__(self, type, value, traceback):
        self.lm._event_queue = None

def load(guidance_file):
    ''' Load a guidance program from the given text file.

    If the passed file is a valid local file it will be loaded directly.
    Otherwise, if it starts with "http://" or "https://" it will be loaded
    from the web.
    '''

    if os.path.exists(guidance_file):
        with open(guidance_file, 'r') as f:
            return f.read()
    elif guidance_file.startswith('http://') or guidance_file.startswith('https://'):
        return requests.get(guidance_file).text
    else:
        raise ValueError('Invalid guidance file: %s' % guidance_file)


def chain(programs, **kwargs):
    ''' Chain together multiple programs into a single program.
    
    This merges them into a single program like: {{>program1 hidden=True}}{{>program2 hidden=True}}
    '''

    from ._program import Program

    new_template = "".join(["{{>program%d hidden=True}}" % i for i in range(len(programs))])
    for i, program in enumerate(programs):
        if isinstance(program, Program):
            kwargs["program%d" % i] = program
        else:
            sig = inspect.signature(program)
            args = ""
            for name, _ in sig.parameters.items():
                args += f" {name}={name}"
            fname = find_func_name(program, kwargs)
            kwargs["program%d" % i] = Program("{{set (%s%s)}}" % (fname, args), **{fname: program})
            # kwargs.update({f"func{i}": program})
    return Program(new_template, **kwargs)


def find_func_name(f, used_names):
    if hasattr(f, "__name__"):
        prefix = f.__name__.replace("<", "").replace(">", "")
    else:
        prefix = "function"

    if prefix not in used_names:
        return prefix
    else:
        for i in range(100):
            fname = f"{prefix}{i}"
            if fname not in used_names:
                return fname


def strip_markers(s):
    """This strips out the comment markers used by guidance."""
    if s is None:
        return None
    return re.sub(r"{{!--G.*?--}}", r"", s, flags=re.MULTILINE | re.DOTALL)

class AsyncIter():    
    def __init__(self, items):    
        self.items = items    

    async def __aiter__(self):    
        for item in self.items:    
            yield item

class TrieOld(object):
    __slots__ = ('children', 'value', 'match_version', 'match', 'partial_match')

    def __init__(self, strings=None, values=None):
        self.children = {}
        self.value = []
        self.match_version = -1
        self.match = False
        self.partial_match = False

        if strings is not None:
            for i,s in enumerate(strings):
                self.insert(s, None if values is None else values[i])

    def insert(self, s, value):
        if len(s) == 0:
            self.value.append(value)
        else:
            first_char = s[0]
            if first_char not in self.children:
                self.children[first_char] = Trie()
            self.children[first_char].insert(s[1:], value)

    def values(self, prefix):
        if prefix == "":
            return [self.value] + list(itertools.chain.from_iterable(self.children[k].values(prefix) for k in self.children))
        else:
            return self.children[prefix[0]].values(prefix[1:])

    def __setitem__(self, key, value):
        if len(key) == 0:
            self.value = value
        else:
            if key[0] not in self.children:
                self.children[key[0]] = Trie()
            self.children[key[0]].__setitem__(key[1:], value)

    def __contains__(self, key):
        return self.__getitem__(key) is not None

    def __getitem__(self, key):
        if len(key) == 0:
            return self.value
        elif key[0] in self.children:
            self.children[key[0]].__getitem__(key[1:])
        else:
            return None

class Trie(object):
    __slots__ = ('children', 'value', 'match_version', 'match', 'partial_match', 'parent', 'flag')

    def __init__(self, strings=None, values=None, parent=None):
        self.children = {}
        self.value = None
        self.match_version = -1
        self.match = False
        self.partial_match = False
        self.parent = parent
        self.flag = None # a spot for user code to store state

        if strings is not None:
            for i,s in enumerate(strings):
                self.insert(s, None if values is None else values[i])

    def insert(self, s, value):
        if len(s) == 0:
            self.value = value
        else:
            first_char = s[0]
            if first_char not in self.children:
                self.children[first_char] = Trie(parent=self)
            self.children[first_char].insert(s[1:], value)

    def values(self, prefix):
        if prefix == "":
            sub_values = list(itertools.chain.from_iterable(self.children[k].values(prefix) for k in self.children))
            if self.value is not None:
                sub_values.append(self.value)
            return sub_values
        else:
            return self.children[prefix[0]].values(prefix[1:])

    def __setitem__(self, key, value):
        if len(key) == 0:
            self.value = value
        else:
            if key[0] not in self.children:
                self.children[key[0]] = Trie(parent=self)
            self.children[key[0]].__setitem__(key[1:], value)

    def __contains__(self, key):
        return self.__getitem__(key) is not None

    def __getitem__(self, key):
        if len(key) == 0:
            return self
        elif key[0] in self.children:
            return self.children[key[0]].__getitem__(key[1:])
        else:
            return None
        
class ByteTrie(object):
    __slots__ = ('children', 'value', 'match_version', 'match', 'partial_match', 'parent', 'flag', 'log_prob')

    def __init__(self, byte_strings=None, values=None, parent=None):
        self.children = {}
        self.value = None
        self.match_version = -1
        self.match = False
        self.partial_match = False
        self.parent = parent
        self.flag = None # a spot for user code to store state
        self.log_prob = 0

        if byte_strings is not None:
            for i,s in enumerate(byte_strings):
                self.insert(s, None if values is None else values[i])

    def insert(self, s, value):
        if len(s) == 0:
            self.value = value
        else:
            first_byte = s[0:1]
            if first_byte not in self.children:
                self.children[first_byte] = ByteTrie(parent=self)
            self.children[first_byte].insert(s[1:], value)

    # def values(self, prefix):
    #     if prefix == "":
    #         sub_values = list(itertools.chain.from_iterable(self.children[k].values(prefix) for k in self.children))
    #         if self.value is not None:
    #             sub_values.append(self.value)
    #         return sub_values
    #     else:
    #         return self.children[prefix[0]].values(prefix[1:])

    # def __setitem__(self, key, value):
    #     if len(key) == 0:
    #         self.value = value
    #     else:
    #         if key[0] not in self.children:
    #             self.children[key[0]] = Trie(parent=self)
    #         self.children[key[0]].__setitem__(key[1:], value)

    # def __contains__(self, key):
    #     return self.__getitem__(key) is not None

    # def __getitem__(self, key):
    #     if len(key) == 0:
    #         return self
    #     elif key[0] in self.children:
    #         return self.children[key[0]].__getitem__(key[1:])
    #     else:
    #         return None

class ContentCapture:
    def __init__(self, variable_stack, hidden=False):
        self._hidden = hidden
        self._variable_stack = variable_stack
    
    def __enter__(self):
        self._pos = len(self._variable_stack["@raw_prefix"])
        if self._hidden:
            self._variable_stack.push({"@raw_prefix": self._variable_stack["@raw_prefix"]})
        return self

    def __exit__(self, type, value, traceback):
        if self._hidden:
            new_content = str(self)
            self._variable_stack.pop()
            self._variable_stack["@raw_prefix"] += "{{!--GHIDDEN:"+new_content.replace("--}}", "--_END_END")+"--}}"

    def __str__(self):
        return strip_markers(self._variable_stack["@raw_prefix"][self._pos:])
    
    def __iadd__(self, other):
        if other is not None:
            self._variable_stack["@raw_prefix"] += other
        return self
    
    def inplace_replace(self, old, new):
        """Replace all instances of old with new in the captured content."""
        self._variable_stack["@raw_prefix"] = self._variable_stack["@raw_prefix"][:self._pos] + self._variable_stack["@raw_prefix"][self._pos:].replace(old, new)

class JupyterComm():
    def __init__(self, target_id, ipython_handle, callback=None, on_open=None, mode="register"):
        from ipykernel.comm import Comm

        self.target_name = "guidance_interface_target_" + target_id
        # print("TARGET NAME", self.target_name)
        self.callback = callback
        self.jcomm = None
        self.ipython_handle = ipython_handle
        self.addd = 1
        self.send_queue = asyncio.Queue()
        self.open_event = asyncio.Event()
        self.is_open = False
        asyncio.get_event_loop().create_task(self._send_loop())
        if mode == "register":
            # log("REGISTERING", self.target_name)
            # asyncio.get_event_loop().create_task(self._register())
            def comm_opened(comm, open_msg):
                # log("OPENED")
                self.addd = 2
                self.jcomm = comm
                self.is_open = True
                self.jcomm.on_msg(self._fire_callback)
                self.open_event.set()
                self._fire_callback({"content": {"data": {"event": "opened"}}})

            self.ipython_handle.kernel.comm_manager.register_target(self.target_name, comm_opened)
            # get_ipython().kernel.comm_manager.register_target(self.target_name, comm_opened) # noqa: F821
        elif mode == "open":
            # log("OPENING", self.target_name)
            self.jcomm = Comm(target_name=self.target_name)
            self.jcomm.on_msg(self._fire_callback)
            # self._fire_callback({"content": {"data": "opened"}})
        else:
            raise Exception("Passed mode must be either 'open' or 'register'!")

    # async def _register(self):
    #     def comm_opened(comm, open_msg):
    #         #log("OPENED")
    #         self.addd = 2
    #         self.jcomm = comm
    #         self.jcomm.on_msg(self._fire_callback)
    #         self.open_event.set()
    #         self._fire_callback({"content": {"data": {"event": "opened"}}})
    #     get_ipython().kernel.comm_manager.register_target(self.target_name, comm_opened)

    # def send(self, data, wait=False):
    #     self.send_queue.append(data)
    #     if self.jcomm is None:
    #         return
    #     for d in self.send_queue:
    #         self.jcomm.send({"data": json.dumps(d)})
    #     self.send_queue = []

    def clear_send_queue(self):
        while not self.send_queue.empty():
            self.send_queue.get_nowait()
            self.send_queue.task_done()

    def _fire_callback(self, msg):
        self.callback(msg["content"]["data"])

    def send(self, data):
        self.send_queue.put_nowait(data)

    async def _send_loop(self):
        while True:
            # log("SENDING_LOOP")
            if self.jcomm is None:
                self.open_event.clear()
                await self.open_event.wait()
            data = await self.send_queue.get()
            # log("SENDING_LOOP got one!")
            self.jcomm.send({"data": json.dumps(data)})

    # async def _waiting_send(self, data):
    #     #log("SENDING", self.jcomm, data)

    #     # await the open event if needed
    #     if self.jcomm is None:
    #         self.open_event.clear()
    #         await self.open_event.wait()
    #     #log("SENDING_now", self.jcomm, data)
    #     self.jcomm.send({"data": json.dumps(data)}) # we encode the JSON so iPython doesn't mess it up


# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')


def escape_template_block(text):
    return text.replace("$", "&#36;").replace("{", "&#123;").replace("}", "&#125;")


def unescape_template_block(text):
    return text.replace("&#36;", "$").replace("&#123;", "{").replace("&#125;", "}")



try:
    import torch
except ImportError:
    pass
import numpy as np
from .._utils import ByteTrie
from ._model import Model
# from ..library._string import string
from .._parser import EarleyCommitParser
from .._grammar import Terminal, select
# import numba

class Local(Model):
    def __init__(self, tokens, bos_token_id, eos_token_id=None, echo=True):
        super().__init__(echo)
        
        assert isinstance(tokens[0], bytes), "The tokens need to be provided as bytes!"

        self.tokens = tokens
        self.bos_token_id = bos_token_id
        self.bos_token = None if self.bos_token_id is None else self.tokens[self.bos_token_id]
        self.eos_token_id = eos_token_id if eos_token_id is not None else bos_token_id
        self.eos_token = None if self.eos_token_id is None else self.tokens[self.eos_token_id]

        # build a prefix tree of the tokens
        self._token_trie = ByteTrie(tokens, np.arange(len(tokens)))
        self._token_trie.match = True
        self._token_trie.match_version = 0

        self._max_token_bytes = max([len(t) for t in self.tokens])  

    def _get_logits(self, token_ids, forced_bytes):
        '''A fake method designed to be overriden by subclasses.'''

        # pretend to extend the KV cache and update the log probs
        return torch.randn(len(self.tokens))

    # def _longest_token_match(self, bytes):
    #     '''Greedy token matching.'''
    #     trie_pos = self._token_trie
    #     for i,c in enumerate(bytes):
    #         if c in trie_pos.children:
    #             trie_pos = trie_pos.children[c]
    #         else:
    #             return bytes[:i], trie_pos.value # note that if there are redudant tokens we choose the one stored in the trie
    #     if len(trie_pos.children) == 0:
    #         return bytes[:i+1], trie_pos.value
    #     else:
    #         return None,None # more than one token can match these bytes
        
    def _tokenize_prefix(self, byte_string):
        '''This is used to speed up the tokenization of long prompts without using the parser.'''
        token_ids = []
        token_byte_positions = []
        
        # loop trying to decode a new token at each iteration
        pos = 0
        while True:

            # walk down the token trie looking for a unique token match
            trie = self._token_trie
            valid_pos = -1
            valid_value = -1
            while True:
                if pos >= len(byte_string):
                    if len(trie.children) > 0:
                        valid_pos = -1
                    break

                # check if we can keep going or are at a dead end
                if byte_string[pos:pos+1] in trie.children:
                    trie = trie.children[byte_string[pos:pos+1]]
                    pos += 1

                    # record the last valid token down this path as we go
                    if trie.value is not None:
                        valid_pos = pos
                        valid_value = trie.value
                else:
                    break # we can't go any farther
            
            if valid_pos == -1:
                break
            else:
                token_ids.append(valid_value)
                token_byte_positions.append(valid_pos)
                pos = valid_pos

        return token_ids,token_byte_positions

    def __call__(self, grammar, max_tokens=100, n=1, top_p=1, temperature=0.0, ensure_bos_token=True, log_probs=False):
        assert n == 1, "Still need to add support for n > 1!"
        
        # get our current context in bytes
        prompt = self._current_prompt()
        prompt = bytes(prompt, encoding="utf8")

        # add the beginning of sequence token if needed
        if ensure_bos_token and self.bos_token is not None and not prompt.startswith(self.bos_token):
            prompt = self.bos_token + prompt
        
        # run a simple tokenizer (that does not use a grammar) on the prefix for better performance
        token_ids,token_byte_positions = self._tokenize_prefix(prompt)
        if len(token_byte_positions) > 0:
            pre_parser_bytes = token_byte_positions[-1]
            prompt = prompt[token_byte_positions[-1]:]
        else:
            pre_parser_bytes = 0
        
        # create a parser with a grammar that includes both our context and the passed grammar
        parser = EarleyCommitParser(prompt + grammar)

        # loop until we have generated a complete pattern
        hidden_count = len(prompt) # we don't emit the prompt
        generated_pos = 0 
        sampled_token_ind = None
        token_count = 0
        last_token_count = 0
        while True: # each iteration generates one more token (and some of the associated bytes)

            # enforce the token limit
            if token_count >= max_tokens:
                break

            # note where we are starting for this token
            start_pos = parser.pos

            # let the parser know that we have advanced another token (used ofr tracking max token limits)
            parser.mark_new_token()

            # walk down the trie as far as possible before computing the logits
            retry_token_gen = False
            trie = self._token_trie
            trie.match_version += 1 # this invalidates all the match caches from the previous token
            while True:
                next_byte_mask = parser.next_byte_mask()
                next_byte_mask_sum = next_byte_mask.sum()
                
                # see if we reached a dead end of the grammar
                if next_byte_mask_sum == 0:
                    break
                
                # if there is more than one option we cannot advance without computing the logits 
                elif next_byte_mask_sum != 1:
                    break

                # we are not forced if we are at the end of the grammar
                elif parser.matched():
                    break

                # if there is only one possible next byte we can keep forcing
                elif next_byte_mask_sum == 1:

                    # look for valid children
                    next_byte = None
                    for byte in trie.children:
                        
                        # mark this trie node with an up-to-date match flag (may save work later)
                        node = trie.children[byte]
                        node.match_version = self._token_trie.match_version
                        node.match = next_byte_mask[byte[0]]
                        
                        # see if we found a match
                        if node.match:
                            next_byte = byte
                            break

                    # if we can't extend then this token is forced
                    if next_byte is None:
                        break
                    
                    # otherwise since there is only one possible next byte we keep going
                    else:
                        commit_point = parser.consume_byte(next_byte, log_prob=0.0)
                        
                        # if we are at a hidden commit point then we need to hide the bytes that match that node
                        if commit_point is not None and commit_point.node.hidden:

                            # This takes the item and commits to it as part of the parse and then shrinks it to zero width
                            # in other words this hides the item
                            parser.commit_and_collapse_item(commit_point)
                            
                            # keep the bytes we still need to emit
                            if start_pos < commit_point.start:
                                parser.shadow_rewind(start_pos)
                            
                            else:
                                # pop off any tokens that overlap the hidden bytes
                                i = len(token_byte_positions) - 1
                                while i >= 0 and token_byte_positions[i] - pre_parser_bytes > commit_point.start:
                                    token_ids.pop()
                                    token_byte_positions.pop()
                                    token_count -= 1
                                    i -= 1
                                # re-add any bytes we cut too far on
                                parser.shadow_rewind(token_byte_positions[-1] - pre_parser_bytes)
                            retry_token_gen = True # this restarts us at the top of the outer token gen loop
                            break
                        
                        trie = trie.children[next_byte]

                
            forced_pos = parser.pos # record how far the bytes are forced

            if retry_token_gen:
                continue

            # back up if we got forced up to a point that is not a valid token
            if next_byte_mask_sum <= 1:
                while trie.value is None and trie.parent is not None:
                    trie = trie.parent
                    forced_pos -= 1
                parser.pos = forced_pos
            
            # if we walked all the way to a forced token then we advance without computing the logits
            # we are forced if there are no more options and we are either in the middle of the grammar or at a trie leaf
            is_forced = next_byte_mask_sum <= 1 and (len(trie.children) == 0 if parser.matched() else trie != self._token_trie)
            if is_forced:
                sampled_token_ind = trie.value
                sampled_token = self.tokens[sampled_token_ind]
                new_bytes_log_prob = 0.0

            # we are at the end of the grammar
            elif next_byte_mask_sum == 0:
                token_pos = 0

                # mark the token we "sampled" if we have comsumed some bytes
                if trie != self._token_trie:
                    sampled_token_ind = trie.value
                    sampled_token = self.tokens[sampled_token_ind]
                    new_bytes_log_prob = 0.0
                    
            # otherwise we need to compute the logits and sample a valid token
            else:
                logits = self._get_logits(token_ids, parser.bytes[start_pos:forced_pos])

                # if requested we compute the log probabilities so we can track the probabilities of each node
                # TODO: we should lower this step to C++ with pybind11
                if log_probs:
                    _compute_log_probs(trie, torch.nn.functional.log_softmax(logits, dim=-1).cpu().numpy())

                # get the sampling order
                grammar_temp = parser.next_byte_temperature()
                current_temp = grammar_temp if grammar_temp >= 0 else temperature # we prefer to use the grammar temp when it is specified
                if current_temp == 0:
                    sampling_order = torch.argsort(-logits, descending=False).cpu().numpy() # we need numpy so the enumerate below does not get really slow...
                else:
                    assert top_p == 1, "Still need to add support for top_p!"
                    probs = torch.nn.functional.softmax(logits / current_temp, dim=-1)
                    sampling_order = torch.multinomial(probs, len(probs)).cpu().numpy()

                # loop over the tokens looking for a valid one
                for i,sampled_token_ind in enumerate(sampling_order):
                    sampled_token = self.tokens[sampled_token_ind]

                    # make sure the parse is backed up to the position we want to start checking from TODO: make this account for shared prefixes with the last token
                    parser.pos = forced_pos
                    new_bytes_log_prob = 0.0

                    # make sure it matches any forced prefix
                    if start_pos < forced_pos and not sampled_token.startswith(parser.bytes[start_pos:forced_pos]):
                        continue
                    offset = forced_pos - start_pos

                    # check to see if the sampled token is allowed
                    token_pos = offset
                    node = trie # this is the Trie node we were left at when we could force the next byte above

                    while token_pos < len(sampled_token):
                        next_byte = sampled_token[token_pos:token_pos+1]
                        next_node = node.children[next_byte]

                        # if we don't have a cached match flag compute it using the grammar
                        if next_node.match_version < self._token_trie.match_version:
                            next_byte_mask = parser.next_byte_mask()
                            for byte in node.children: # we update all the children since the parser knows the full mask
                                child = node.children[byte]
                                child.match_version = self._token_trie.match_version
                                child.match = next_byte_mask[byte[0]]
                        
                        # advance or fail according to the (now up-to-date) match cache
                        if next_node.match:
                            
                            # mark that we accepted this byte
                            node = next_node
                            token_pos += 1

                            # get the parser to consume the next byte
                            log_prob_delta = next_node.log_prob - node.log_prob
                            new_bytes_log_prob += log_prob_delta
                            commit_point = parser.consume_byte(next_byte, log_prob=log_prob_delta)
                        
                            # if we are at a hidden commit point then we need to hide the bytes that match that node
                            if commit_point is not None and commit_point.node.hidden:

                                # This takes the item and commits to it as part of the parse and then shrinks it to zero width
                                # in other words this hides the item
                                parser.commit_and_collapse_item(commit_point)
                                
                                # keep the bytes we still need to emit
                                if forced_pos < commit_point.start:
                                    parser.shadow_rewind(forced_pos)
                                
                                else:
                                    # pop off any tokens that overlap the hidden bytes
                                    i = len(token_byte_positions) - 1
                                    while i >= 0 and token_byte_positions[i] - pre_parser_bytes > commit_point.start:
                                        token_ids.pop()
                                        token_byte_positions.pop()
                                        token_count -= 1
                                        i -= 1
                                    # re-add any bytes we cut too far on
                                    parser.shadow_rewind(token_byte_positions[-1] - pre_parser_bytes)
                                retry_token_gen = True # this restarts us at the top of the outer token gen loop
                                break

                            elif token_pos == len(sampled_token):
                                break # this token is valid
                        else:
                            # partially valid tokens are okay if we are running off the end of a grammar, but not otherwise
                            if not parser.matched():
                                token_pos = -1

                            break # this token is no longer valid

                    # see if we are breaking out of the whole loop
                    if retry_token_gen:
                        break

                    # check if this token is dominated by other longer valid tokens (and hence would never be consistent with greedy tokenization)
                    if token_pos == len(sampled_token) and not parser.matched(): # not we don't check if we have matched, because then we can generate anything afterwards
                        if _check_dominated(node, parser, self._token_trie.match_version, parser.next_byte_mask()):
                            token_pos = -1

                    if token_pos > 0:
                        break # we found a valid token

                    if parser.matched():
                        break # if we already have a full match we don't try more tokens we just give up as soon as the model deviates from the grammar
            
            # if we just collpased a hidden commit point then we start over looking for a new token
            if retry_token_gen:
                continue

            # emit whatever we know will not be hidden
            new_bytes = parser.bytes[generated_pos:parser.earliest_hidden_start()]

            # if we cannot consume any more tokens then we are done
            if not is_forced and token_pos < len(sampled_token) and trie == self._token_trie:
                assert parser.matched(), "We can't consume any more tokens, but we are not yet done! Perhaps your model's token set is incomplete?"

                # TODO: if we exactly match the end of the pattern then we can commit to this last token 
                # if m.span()[1] == len(generated_text):
                #     self._cache_state["new_token_ids"].append(sampled_token_ind)

                # capture the named groups from the parse tree
                parse_tree = parser.parse_tree()
                data = {}
                log_prob_data = {}
                _record_captures(parse_tree, data, log_prob_data, parser.bytes)
                
                # we have no valid log prob data if we didn't compute it
                if not log_probs:
                    log_prob_data = {k: None for k in data}
                yield new_bytes[hidden_count:], not is_forced, new_bytes_log_prob, data, log_prob_data, token_count - last_token_count
                last_token_count = token_count
                break # we are done!
            else:
                generated_pos += len(new_bytes)

                # yeild the snippet of text created by the next token
                out = new_bytes[hidden_count:]
                if len(out) > 0:
                    yield out, not is_forced, new_bytes_log_prob, {}, {}, token_count - last_token_count # note that we don't capture groups until a complete parse right now...
                    last_token_count = token_count
                    hidden_count = 0
                    token_count += 1 # note we only update this for tokens that emit non-hidden content
                else:
                    hidden_count -= len(new_bytes)

                token_ids.append(sampled_token_ind)

                # track the byte position of each token
                if len(token_byte_positions) == 0:
                    token_byte_positions.append(len(sampled_token))
                else:
                    token_byte_positions.append(token_byte_positions[-1] + len(sampled_token))

def _record_captures(initial_item, data, log_prob_data, byte_data):
    stack = [(initial_item, 0)]
    used_names = set() # track which capture names have been used so self-recursive children don't overwrite their parents
    
    while stack:
        item, byte_pos = stack.pop()
        # terminal nodes
        if isinstance(item, Terminal):

            # if we are at a capture group node then we save the matched terminal byte
            if item.capture_name is not None:
                data[item.capture_name] = item.byte
                log_prob_data[item.capture_name] = 0
        
        # internal nodes
        else:
            start_byte_pos = byte_pos

            # recurse for all our non-null children
            for child in item.children:
                if child is not None:
                    stack.append((child, byte_pos))
                    # _record_captures(child, data, log_prob_data, byte_data, byte_pos)
                    if isinstance(child, Terminal):
                        byte_pos += len(child)
                    else:
                        byte_pos = child.start # note that "start" means "end" since this is a reversed state set

            # if we are at a capture group node then we save the matched bytes range
            # note that we record this after calling our children so that we save the outermost version of self-recursive calls
            cname = item.node.capture_name
            if cname is not None and cname not in used_names:
                
                # see if we are doing a list append
                if cname.startswith("__LIST_APPEND:"):
                    cname = cname[14:] # trim off the list append tag
                    if cname not in data or not isinstance(data[cname], list):
                        data[cname] = []
                        log_prob_data[cname] = []
                    data[cname].append(byte_data[start_byte_pos:item.start])
                    log_prob_data[cname].append(item.log_prob)
                
                # or just a regular assignment
                else:
                    data[cname] = byte_data[start_byte_pos:item.start] # note that "start" means "end" since this is a reversed state set
                    log_prob_data[cname] = item.log_prob

                used_names.add(cname)    

def _compute_log_probs(trie, log_probs):
    '''Computes the log probabilities for each internal trie node.'''
    if trie.value is not None:
        trie.log_prob += log_probs[trie.value]
    
    if len(trie.children) > 0:
        child_log_probs = []
        for b in trie.children:
            child = trie.children[b]
            _compute_log_probs(child, log_probs)
            child_log_probs.append(child.log_prob)
        trie.log_prob = np.logaddexp.reduce(child_log_probs)

def _check_dominated(node, parser, match_version, next_byte_mask):
    curr_pos = parser.pos
    for byte_num in next_byte_mask.nonzero()[0]:
        next_byte = bytes((byte_num,))
        if next_byte not in node.children:
            return False # no possible exension this direction, so we are not dominated
        child = node.children[next_byte]
        if child.match_version < match_version:
            child.match_version = match_version
            child.match = next_byte_mask[next_byte[0]]
        
        if not child.match:
            return False # this child does not dominate the node, so the node is not dominated
        elif child.value is None: # this child might not dominate the node
            parser.consume_byte(next_byte, log_prob=0.0)
            child_dominate = _check_dominated(child, parser, match_version, parser.next_byte_mask())
            parser.pos = curr_pos
            if not child_dominate:
                return False
    return True



import guidance
from .._grammar import select, string

@guidance(stateless=True, dedent=False)
def prefix_tree(lm, strings, partial_matches=False):

    if len(strings) == 0:
        return lm

    # group the strings by their starting character
    char_groups = {}
    for s in strings:
        if len(s) > 0:
            if s[0] not in char_groups:
                char_groups[s[0]] = []
            char_groups[s[0]].append(s[1:])
    
    # enable any empty followup if partial matches are allowed
    if partial_matches:
        char_groups[""] = []
    
    # recursively build the tree
    suboptions = [string(k) + prefix_tree(v, partial_matches=partial_matches) for k,v in char_groups.items()]

    return lm + select(suboptions, skip_checks=True) # we skip normal type checks for speed


import guidance
from .._grammar import byte_range

@guidance(stateless=True)
def any_char(lm):
    # TODO: extend this to support utf-8 encoded multibyte unicode characters
    return lm + byte_range(b'\x00', b'\xff')


from guidance import models, char_set

def test_single_char():
    model = models.LocalMock("<s>abc")
    assert str(model + '<s>' + char_set("a")) == "<s>a"
    assert str(model + '<s>' + char_set("ab")) == "<s>a"
    assert str(model + '<s>' + char_set("ba")) == "<s>a"
    assert str(model + '<s>' + char_set("b")) == "<s>b"

def test_char_range():
    model = models.LocalMock("<s>bac")
    assert str(model + '<s>' + char_set("a-c")) == "<s>b"
    assert str(model + '<s>' + char_set("b-z")) == "<s>b"
    assert str(model + '<s>' + char_set("0-9")) != "<s>b"
    assert str(model + '<s>' + char_set("b0-9")) == "<s>b"


import os
from pathlib import Path
import multiprocessing
from itertools import takewhile
import operator
import threading
import numpy as np
import queue
import time
import tiktoken

from ._vertexai import VertexAIInstruct, VertexAIChat

# try:
#     # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
#     import torch
#     is_torch = True
# except ImportError:
#     is_torch = False

try:
    # TODO: can we eliminate the torch requirement for llama.cpp by using numpy in the caller instead?
    from vertexai.preview.language_models import TextGenerationModel
    from vertexai.language_models import ChatModel, InputOutputTextPair
    is_vertexai = True
except ImportError:
    is_vertexai = False

class PaLM2Instruct(VertexAIInstruct):
    def __init__(self, model, tokenizer=None, echo=True, caching=True, temperature=0.0, max_streaming_tokens=500, **kwargs):\
    
        if isinstance(model, str):
            self.model_name = model
            self.model_obj = TextGenerationModel.from_pretrained(self.model_name)
        
        # PaLM2 does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")
        
        # the superclass does all the work
        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            caching=caching,
            temperature=temperature,
            max_streaming_tokens=max_streaming_tokens,
            **kwargs
        )

class PaLM2Chat(VertexAIChat):
    def __init__(self, model, tokenizer=None, echo=True, caching=True, temperature=0.0, max_streaming_tokens=500, **kwargs):\
    
        if isinstance(model, str):
            self.model_name = model
            self.model_obj = ChatModel.from_pretrained(self.model_name)
        
        # PaLM2 does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")
        
        # the superclass does all the work
        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            caching=caching,
            temperature=temperature,
            max_streaming_tokens=max_streaming_tokens,
            **kwargs
        )

    # def get_role_start(self, role_name, **kwargs):
    #     if role_name == "user":

    #         # if we follow an auto-nested system role then we are done
    #         if self._current_prompt().endswith("\n<</SYS>>\n\n"):
    #             return ""
    #         else:
    #             return "[INST] "
        
    #     elif role_name == "assistant":
    #         return " "
        
    #     elif role_name == "system":
            
    #         # check if we are already embedded at the top of a user role
    #         if self._current_prompt().endswith("[INST] "):
    #             return "<<SYS>>\n"

    #         # if not then we auto nest ourselves
    #         else:
    #             return "[INST] <<SYS>>\n"
    
    # def get_role_end(self, role_name=None):
    #     if role_name == "user":
    #         return " [/INST]"
    #     elif role_name == "assistant":
    #         return " "
    #     elif role_name == "system":
    #         return "\n<</SYS>>\n\n"


import guidance
from guidance import select, gen
from ..utils import get_model

def test_llama_cpp_gen():
    lm = get_model("llama_cpp:")
    lm = lm + "this is a test" + gen("test", max_tokens=10)
    assert len(str(lm)) > len("this is a test")

def test_llama_cpp_recursion_error():
    lm = get_model("llama_cpp:")

    # define a guidance program that adapts a proverb
    lm = lm + f"""Tweak this proverb to apply to model instructions instead.
    {gen('verse', max_tokens=2)}
    """
    assert len(str(lm)) > len("Tweak this proverb to apply to model instructions instead.\n\n")

def test_llama_cpp_select2():
    lm = get_model("llama_cpp:")
    lm += f'this is a test1 {select(["item1", "item2"])} and test2 {select(["item3", "item4"])}'
    assert str(lm) in [
        "this is a test1 item1 and test2 item3", 
        "this is a test1 item1 and test2 item4",
        "this is a test1 item2 and test2 item3", 
        "this is a test1 item2 and test2 item4"]

def test_repeat_calls():
    llama2 = get_model("llama_cpp:")
    a = []
    lm = llama2 + 'How much is 2 + 2? ' + gen(name='test', max_tokens=10)
    a.append(lm['test'])
    lm = llama2 + 'How much is 2 + 2? ' + gen(name='test',max_tokens=10, pattern='\d+')
    a.append(lm['test'])
    lm = llama2 + 'How much is 2 + 2? ' + gen(name='test', max_tokens=10)
    a.append(lm['test'])
    assert a[-1] == a[0]

def test_suffix():
    llama2 = get_model("llama_cpp:")
    lm = llama2 + '1. Here is a sentence ' + gen(name='bla', list_append=True, suffix='\n')
    assert (str(lm))[-1] == '\n'
    assert (str(lm))[-2] != '\n'

def test_subtoken_forced():
    llama2 = get_model("llama_cpp:")
    lm = llama2 + 'How much is 2 + 2? ' + gen(name='test', max_tokens=10, regex='\(')
    assert str(lm) == "How much is 2 + 2? ("



from typing import Any
try:
    from IPython.display import clear_output, display, HTML
except ImportError:
    clear_output = lambda wait=True: None
    display = lambda arg: None
    HTML = lambda arg: None
import html
import re
import copy
import json
import textwrap
import numpy as np
from .._grammar import StatelessFunction, StatefulFunction, tag_start, tag_end, string, _call_pool, _tag_pattern, Null, replace_model_variables, unreplace_model_variables, commit_point, select

class Endpoint:
    '''This keeps state that is shared among all models using the same endpoint session.'''
    pass

_null_grammar = string('')
format_pattern = re.compile(r"<\|\|_.*?_\|\|>", flags=re.DOTALL)
nodisp_pattern = re.compile(r"&lt;\|\|_#NODISP_\|\|&gt;.*?&lt;\|\|_/NODISP_\|\|&gt;", flags=re.DOTALL)
html_pattern = re.compile(r"&lt;\|\|_html:(.*?)_\|\|&gt;", flags=re.DOTALL)

class Model:
    _open_blocks = {}
    _context_free_mode = 0

    def __init__(self, echo=True):

        # while models are logically immutable, they can share some mutable caching state to save computation
        self._cache_state  = {}
        self.echo = echo
        self._state = ""
        self._children = []
        self._opened_blocks = {}
        self._event_queue = None
        self._event_parent = None
        self._silent = None
        self._token_count = 0
        self._variables = {}
        self.instance__enter__ = []
        self.instance__exit__ = []
        self._streaming = False

    def __call__(self, pattern=None, max_tokens=100, n=1, top_p=1, temperature=0.0, ensure_bos_token=True):
        pass # meant to be overriden by subclasses

    @property
    def default_end_patterns(self):

        # add the eos token
        parts = [self.eos_token]

        # add any active non empty role ends. Ignore role ends that are spaces
        for role_end_str in self._opened_blocks.values():
            role_end_str = format_pattern.sub("", role_end_str)
            if len(role_end_str) > 0 and not re.fullmatch(r'\s+', role_end_str):
                parts.append(role_end_str)

        return select(parts)

    def get(self, key, default=None):
        return self._variables.get(key, default)

    def set(self, key, value):
        copy = self.copy()
        copy._variables[key] = value
        return copy
    
    def remove(self, key):
        if key in self._variables:
            del self._variables[key]

    def _html(self):
        display_out = self._state
        for context in reversed(self._opened_blocks):
            display_out += self._opened_blocks[context]
        display_out = html.escape(display_out)
        display_out = nodisp_pattern.sub("", display_out)
        display_out = html_pattern.sub(lambda x: html.unescape(x.group(1)), display_out)
        display_out = "<pre style='margin: 0px; padding: 0px; padding-left: 8px; margin-left: -8px; border-radius: 0px; border-left: 1px solid rgba(127, 127, 127, 0.2); white-space: pre-wrap; font-family: ColfaxAI, Arial; font-size: 15px; line-height: 23px;'>"+display_out+"</pre>"
        return display_out
    
    def _send_to_event_queue(self, value):
        if self._event_queue is not None:
            self._event_queue.put(value)
        if self._event_parent is not None:
            self._event_parent._send_to_event_queue(value)

    def is_silent(self):
        if self._silent is not None:
            return self._silent
        return False
    
    def copy(self):
        
        # start with a shallow copy
        new_lm = copy.copy(self)

        # then copy a few things we need deeper copies of
        new_lm._variables = self._variables.copy()
        new_lm._opened_blocks = self._opened_blocks.copy()
        
        # create a new clean event queue # TODO: can we delete this now?
        new_lm._event_queue = None
        if self._event_queue is not None:
            new_lm._event_parent = self
        
        # track parent/child relationships
        new_lm._children = []
        self._children.append(new_lm)
        
        return new_lm
    
    def _inplace_append(self, value, force_silent=False):
        """This is the base way to add content to the LM object."""
        self._state += str(value)
        if self.echo and not self.is_silent() and not force_silent:
            clear_output(wait=True)
            display(HTML(self._html()))
        self._send_to_event_queue(self)
        return self
    
    def reset(self, clear_variables=True):
        """This resets the state of the LM prompt."""
        self._reset(0, clear_variables)
        return self

    def _reset(self, position=0, clear_variables=True):
        self._state = self._state[:position]
        if clear_variables:
            self._variables = {}

    def _repr_html_(self):
        clear_output(wait=True)
        return self._html()
    
    def _current_prompt(self):
        return format_pattern.sub("", self._state)
    
    def __str__(self) -> str:
        out = self._current_prompt()
        for context in reversed(self._opened_blocks):
            out += format_pattern.sub("", self._opened_blocks[context])
        return out
    
    def __add__(self, value):

        # create the new lm object we will return
        # (we need to do this since Model objects are immutable)
        lm = self.copy()

        # close any newly closed contexts
        for context in list(reversed(lm._opened_blocks)):
            if context not in Model._open_blocks and context in lm._opened_blocks:
                close_text = lm._opened_blocks[context] # save so we can delete it before adding it
                del lm._opened_blocks[context]
                lm._inplace_append(close_text)

        # apply any newly opened contexts (new from this object's perspective)
        for context in Model._open_blocks:
            if context not in lm._opened_blocks:
                lm._opened_blocks[context] = "" # mark this so we don't readd when computing the opener (even though we don't know the close text yet)
                lm += context.opener
                with context_free():
                    tmp = lm + context.closer
                close_text = tmp._state[len(lm._state):] # get the new state added by calling the closer
                lm._opened_blocks[context] = close_text
        
        # wrap raw string values
        if isinstance(value, str):
            is_id = False
            parts = re.split(_tag_pattern, value)
            
            # we have no embedded objects
            if len(parts) == 1:
                return lm._inplace_append(value)
            
            # if we have embedded objects we have to convert the string to a grammar tree
            partial_grammar = _null_grammar
            lm.suffix = ""
            for i,part in enumerate(parts):
                if i < len(parts) - 1:
                    lm.suffix = parts[i+1]
                if is_id:
                    call = _call_pool[part]
                    if isinstance(call, StatelessFunction):
                        partial_grammar += _call_pool[part]
                    else:
                        lm += partial_grammar
                        lm = _call_pool[part](lm)
                        partial_grammar = _null_grammar
                elif part != "":
                    partial_grammar += string(part)
                is_id = not is_id
            return lm + partial_grammar
        
        elif isinstance(value, Null):
            return lm
        
        # run stateless functions (grammar nodes)
        elif isinstance(value, StatelessFunction):
            return lm.run_stateless(value)
        
        # run stateful functions
        else:
            return value(lm)
    
    def endswith(self, s):
        return self._state.endswith(s)
    
    def __len__(self):
        return len(str(self))
    
    def __setitem__(self, key, value):
        self._variables[key] = value

    def __getitem__(self, key):
        return self._variables[key]
    
    def __contains__(self, item):
        return item in self._variables

    # def __enter__(self):
    #     Model._open_blocks
    #     self._opened_blocks
    #     if len(self.instance__enter__) > 0:
    #         return self.instance__enter__.pop()()

    # def __exit__(self, exc_type, exc_value, traceback):
    #     if len(self.instance__exit__) > 0:
    #         return self.instance__exit__.pop()(exc_type, exc_value, traceback)
    
    def get_cache(self):
        return self.engine.cache
    
    def tool_def(self, functions):

        self += """
# Tools

"""
        if len(functions) > 0:
            self += '''## functions

namespace functions {

'''
        for function in functions:
            self += f"""// {function['description']}
type {function['name']} = (_: {{"""
            for prop_name,prop_data in function["parameters"]["properties"].items():
                if "description" in prop_data:
                    self += f"\n// {prop_data['description']}\n"
                self += prop_name
                if prop_name not in function["parameters"]["required"]:
                    self += "?"
                self += ": "
                if "enum" in prop_data:
                    for enum in prop_data["enum"]:
                        self += f'"{enum}"'
                        if enum != prop_data["enum"][-1]:
                            self += " | "
                else:
                    self += prop_data["type"]
                
                if prop_name != list(function["parameters"]["properties"].keys())[-1]:
                    self += ",\n"
            self += """
}) => any;

"""
            self[function['name']] = function
        self += "} // namespace functions\n"
        
        return self

    def run_stateless(lm, stateless_function, max_tokens=1000, temperature=0.0, top_p=1.0, n=1):
        assert Model._context_free_mode == 0, "We can't run grammar parsing while in context free mode! (for example inside a block closer)"

        # This needs to be here for streaming
        # if name is not None:
        #     lm[name] = ""


        # replace ModelVariables with their actual values (note we save what we replaced so we can restore it later)
        replacements = replace_model_variables(stateless_function, lm)

        # start the generation stream
        gen_obj = lm(
            grammar=stateless_function, max_tokens=max_tokens, n=n,
            temperature=temperature, top_p=top_p
        )

        # single generation
        if n == 1:
            generated_value = ""
            # logprobs_out = []

            delayed_bytes = b""
            # last_is_generated = False
            for new_bytes, is_generated, new_bytes_log_prob, capture_groups, capture_group_log_probs, new_token_count in gen_obj:
                # convert the bytes to a string (delaying if we don't yet have a valid unicode string)
                lm._token_count += new_token_count
                new_bytes = delayed_bytes + new_bytes
                try:
                    new_text = new_bytes.decode("utf8")
                except UnicodeDecodeError:
                    delayed_bytes = new_bytes
                    continue
                delayed_bytes = b""

                if len(new_bytes) > 0:
                    generated_value += new_text
                    if is_generated:
                        lm += f"<||_html:<span style='background-color: rgba(0, 165, 0, {0.15 + 0.4 * (1 - np.exp(new_bytes_log_prob))}); border-radius: 3px;' title='{new_bytes_log_prob}'>_||>"
                    lm += new_text
                    if is_generated:
                        lm += "<||_html:</span>_||>"
                
                # last_is_generated = is_generated

                if len(capture_groups) > 0:
                    for k in capture_groups:
                        v = capture_groups[k]
                            
                        # see if we are in a list_append mode
                        if isinstance(v, list):
                            for inner_v in v:
                                # convert to a string if possible
                                # TODO: will need to not just always do this once we support images etc.
                                try:
                                    inner_v = inner_v.decode("utf8") if isinstance(inner_v, bytes) else inner_v
                                except UnicodeDecodeError:
                                    pass

                                if k not in lm or not isinstance(lm[k], list):
                                    lm[k] = []
                                lm[k].append(inner_v)

                        # ...or standard assignment mode
                        else:
                            # convert to a string if possible
                            # TODO: will need to not just always do this once we support images etc.
                            try:
                                v = v.decode("utf8") if isinstance(v, bytes) else v
                            except UnicodeDecodeError:
                                pass
                            lm[k] = v

            # if len(capture_groups) > 0:
            #     for k in capture_groups:
            #         v = capture_groups[k]
            #         lm[k] = v.decode("utf8") if isinstance(v, bytes) else v
        
        unreplace_model_variables(replacements)

        return lm

class Chat(Model):
    
    def get_role_start(self, role_name, **kwargs):
        return "<|im_start|>"+role_name+"".join([f' {k}="{v}"' for k,v in kwargs.items()])+"\n"
    
    def get_role_end(self, role_name=None):
        return "<|im_end|>"
    
class Instruct(Model):
    
    def get_role_start(self, role_name, **kwargs):
        return "<|im_start|>"+role_name+"".join([f' {k}="{v}"' for k,v in kwargs.items()])+"\n"
    
    def get_role_end(self, role_name=None):
        return "<|im_end|>"
    
class ContextFree:
    def __enter__(self):
        Model._context_free_mode += 1
    
    def __exit__(self, exc_type, exc_value, traceback):
        Model._context_free_mode -= 1

def context_free():
    return ContextFree()